@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2007, 2009, 2010 Free Software Foundation, Inc.
+   Copyright (C) 2007-2024 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,9 +12,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with GNU Mailutils; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-   MA 02110-1301 USA */
+   along with GNU Mailutils.  If not, see <http://www.gnu.org/licenses/>. */
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
@@ -28,6 +26,7 @@
 char *file;
 mu_header_t header;
 mu_iterator_t iterator;
+mu_stream_t hstream;
 
 char *ps[] = { "> ", ". " };
 int interactive;
@@ -83,7 +82,7 @@ load_file (const char *name)
 
   buf[st.st_size] = '\n';
   buf[st.st_size+1] = 0;
-  status = mu_header_create (&header, buf, st.st_size + 1, NULL);
+  status = mu_header_create (&header, buf, st.st_size + 1);
   free (buf);
   if (status)
     {
@@ -124,7 +123,8 @@ cmd_load (int argc, char **argv)
 {
   if (check_args (argv[0], argc, 2, 2))
     return;
-  mu_header_destroy (&header, NULL);
+  mu_stream_destroy (&hstream);
+  mu_header_destroy (&header);
   load_file (argv[1]);
 }
 
@@ -134,7 +134,8 @@ cmd_free (int argc, char **argv)
   if (check_args (argv[0], argc, 1, 1))
     return;
   mu_iterator_destroy (&iterator);
-  mu_header_destroy (&header, NULL);
+  mu_stream_destroy (&hstream);
+  mu_header_destroy (&header);
 }
 
 void
@@ -173,25 +174,27 @@ cmd_dump (int argc, char **argv)
   if (argc == 2)
     off = strtoul (argv[1], NULL, 0);
 
-  status = mu_header_get_stream (header, &stream);
+  status = mu_header_get_streamref (header, &stream);
   if (status)
     {
       mu_error ("%u: cannot get stream: %s", line_num, mu_strerror (status));
       return;
     }
 
-  status = mu_stream_seek (stream, off, SEEK_SET);
+  status = mu_stream_seek (stream, off, SEEK_SET, NULL);
   if (status)
     {
+      mu_stream_destroy (&stream);
       mu_error ("%u: cannot seek: %s", line_num, mu_strerror (status));
       return;
     }
 
-  while (mu_stream_sequential_read (stream, buf, sizeof buf, &n) == 0
+  while (mu_stream_read (stream, buf, sizeof buf, &n) == 0
 	 && n > 0)
     {
       fwrite (buf, 1, n, stdout);
     }
+  mu_stream_destroy (&stream);
 }  
 
 void
@@ -209,6 +212,7 @@ cmd_remove (int argc, char **argv)
   status = mu_header_remove (header, fn, num);
   if (status)
     mu_error ("%u: %s: %s", line_num, argv[0], mu_strerror (status));
+  mu_stream_destroy (&hstream);
 }
 
 /* insert header value [ref [num] [before|after] [replace]] */
@@ -258,6 +262,7 @@ cmd_insert (int argc, char **argv)
     }
   status = mu_header_insert (header, argv[1], argv[2],
 			     ref, num, flags);
+  mu_stream_destroy (&hstream);
   if (status)
     mu_error ("%u: %s: %s", line_num, argv[0], mu_strerror (status));
 }
@@ -272,20 +277,87 @@ cmd_write (int argc, char **argv)
   if (check_args (argv[0], argc, 1, 1))
     return;
 
-  status = mu_header_get_stream (header, &str);
+  status = mu_header_get_streamref (header, &str);
   if (status)
     {
       mu_error ("%u: cannot get stream: %s", line_num, mu_strerror (status));
       return;
     }
-  printf("[reading headers; end with an empty line]\n");
-  mu_stream_seek (str, 0, SEEK_SET);
+  printf ("[reading headers; end with an empty line]\n");
+  mu_stream_seek (str, 0, SEEK_SET, NULL);
   while (prompt (1), fgets(buf, sizeof buf, stdin))
     {
-      mu_stream_sequential_write (str, buf, strlen (buf));
+      mu_stream_write (str, buf, strlen (buf), NULL);
       if (buf[0] == '\n')
 	break;
     }
+  mu_stream_destroy (&str);
+  mu_stream_destroy (&hstream);
+}
+
+void
+cmd_overwrite (int argc, char **argv)
+{
+  char buf[512];
+  mu_stream_t str;
+  int status;
+  mu_off_t off;
+  
+  if (check_args (argv[0], argc, 2, 2))
+    return;
+
+  off = strtoul (argv[1], NULL, 0);
+  
+  status = mu_header_get_streamref (header, &str);
+  if (status)
+    {
+      mu_error ("%u: cannot get stream: %s", line_num, mu_strerror (status));
+      return;
+    }
+  status = mu_stream_seek (str, off, SEEK_SET, NULL);
+  if (status)
+    {
+      mu_error ("seek error: %s", mu_strerror (status));
+      return;
+    }
+
+  printf ("[reading headers; end with an empty line]\n");
+  while (prompt (1), fgets(buf, sizeof buf, stdin))
+    {
+      if (buf[0] == '\n')
+	break;
+      mu_stream_write (str, buf, strlen (buf), NULL);
+    }
+  mu_stream_destroy (&str);
+  mu_stream_destroy (&hstream);
+}
+
+void
+cmd_append (int argc, char **argv)
+{
+  char buf[512];
+  mu_stream_t str;
+  int status;
+  
+  if (check_args (argv[0], argc, 1, 1))
+    return;
+
+  status = mu_header_get_streamref (header, &str);
+  if (status)
+    {
+      mu_error ("%u: cannot get stream: %s", line_num, mu_strerror (status));
+      return;
+    }
+  printf ("[reading headers; end with an empty line]\n");
+  mu_stream_seek (str, 0, SEEK_END, NULL);
+  while (prompt (1), fgets(buf, sizeof buf, stdin))
+    {
+      mu_stream_write (str, buf, strlen (buf), NULL);
+      if (buf[0] == '\n')
+	break;
+    }
+  mu_stream_destroy (&str);
+  mu_stream_destroy (&hstream);
 }
 
 void
@@ -338,18 +410,19 @@ void
 cmd_readline (int argc, char **argv)
 {
   char *buf;
-  size_t size;
-  mu_stream_t stream;
+  size_t size = 128;
   size_t nbytes;
   
   if (check_args (argv[0], argc, 1, 2))
     return;
-  size = atoi (argv[1]);
+  if (argc == 2)
+    size = atoi (argv[1]);
   buf = malloc (size);
   if (!buf)
     abort ();
-  mu_header_get_stream (header, &stream);
-  mu_stream_readline (stream, buf, size, 0, &nbytes);
+  if (!hstream)
+    mu_header_get_streamref (header, &hstream);
+  mu_stream_readline (hstream, buf, size, &nbytes);
   printf ("\"%*.*s\"", (int) nbytes, (int) nbytes, buf);
   free (buf);
 }
@@ -371,7 +444,7 @@ static struct cmdtab cmdtab[] = {
   { "free", cmd_free, NULL, "discard all headers" },
   { "print", cmd_print, "NAME [N]",
     "find and print the Nth (by default, 1st) instance of header named NAME" },
-  { "dump", cmd_dump, NULL, "dump all headers on screen" },
+  { "dump", cmd_dump, "[OFF]", "dump all headers on screen" },
   { "itr", cmd_iterate, "[first|1|next|n]", "iterate over headers" },
   { "readline", cmd_readline, "[SIZE]", "read line" },
   { "remove", cmd_remove, "NAME [N]",
@@ -379,6 +452,8 @@ static struct cmdtab cmdtab[] = {
   { "insert", cmd_insert, "NAME VALUE [REF [NUM] [before|after] [replace]]",
     "insert new header" },
   { "write", cmd_write, NULL, "accept headers from raw stream" },
+  { "overwrite", cmd_overwrite, "OFF", "overwrite raw data from offset OFF" },
+  { "append", cmd_append, NULL, "append raw data" },
   { "help", cmd_help, "[COMMAND]", "print short usage message" },
   { NULL }
 };
@@ -481,9 +556,9 @@ main (int argc, char **argv)
 {
   int c;
   char buf[512];
-  char **prevv;
-  int prevc = 0;
-  
+  char **prevv = NULL;
+  size_t prevc = 0;
+
   interactive = isatty (0);
   while ((c = getopt (argc, argv, "f:h")) != EOF)
     {
@@ -509,7 +584,7 @@ main (int argc, char **argv)
     }
   else
     {
-      int status = mu_header_create (&header, NULL, 0, NULL);
+      int status = mu_header_create (&header, NULL, 0);
       if (status)
 	{
 	  mu_error ("cannot create header: %s", mu_strerror (status));
@@ -517,36 +592,33 @@ main (int argc, char **argv)
 	}
     }
   
-  while (prompt(0), fgets(buf, sizeof buf, stdin))
+  while (prompt (0), fgets (buf, sizeof buf, stdin))
     {
-      int c;
-      char **v;
-      int status;
+      struct mu_wordsplit ws;
 
       line_num++;
-      status = mu_argcv_get (buf, NULL, "#", &c, &v);
-      if (status)
+      ws.ws_comment = "#";
+      if (mu_wordsplit (buf, &ws, MU_WRDSF_DEFFLAGS | MU_WRDSF_COMMENT))
 	{
-	  mu_error ("%u: cannot parse: %s",
-		   line_num, mu_strerror (status));
+	  mu_error ("cannot split line `%s': %s", buf,
+		    mu_wordsplit_strerror (&ws));
 	  continue;
 	}
-
-      if (c == 0)
+      
+      if (ws.ws_wordc == 0)
 	{
 	  if (prevc)
 	    docmd (prevc, prevv);
-	  else
-	    mu_argcv_free (c, v);
 	}
       else
 	{
-	  docmd (c, v);
+	  docmd (ws.ws_wordc, ws.ws_wordv);
 	  mu_argcv_free (prevc, prevv);
-	  prevc = c;
-	  prevv = v;
+	  mu_wordsplit_get_words (&ws, &prevc, &prevv);
 	}
+      mu_wordsplit_free (&ws);
     }
+  mu_argcv_free (prevc, prevv);
   exit (0);
 }
 
