@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2019-2020 Free Software Foundation, Inc.
+   Copyright (C) 2019-2024 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
@@ -37,6 +37,7 @@
 #include <mailutils/nls.h>
 #include <mailutils/header.h>
 #include <mailutils/attribute.h>
+#include <mailutils/envelope.h>
 #include <mailutils/util.h>
 #include <mailutils/cctype.h>
 
@@ -69,34 +70,17 @@ dotmail_mailbox_init_stream (struct mu_dotmail_mailbox *dmp)
   int rc;
   mu_mailbox_t mailbox = dmp->mailbox;
 
-  /*
-   * Initialize stream flags.  If append mode is requested, convert it to
-   * read-write, so that dotmail_flush_unlocked be able to update the
-   * X-IMAPbase header in the first message, if necessary.
-   */
-  dmp->stream_flags = mailbox->flags;
-  if (dmp->stream_flags & MU_STREAM_APPEND)
-    dmp->stream_flags = (dmp->stream_flags & ~MU_STREAM_APPEND) | MU_STREAM_RDWR;
-  rc = mu_mapfile_stream_create (&mailbox->stream, dmp->name, dmp->stream_flags);
+  rc = mu_mailbox_stream_create (&mailbox->stream, dmp->name, mailbox->flags);
   if (rc)
     {
       mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
 		("%s:%s (%s): %s",
-		 __func__, "mu_mapfile_stream_create", dmp->name,
+		 __func__, "mu_mailbox_file_open", dmp->name,
 		 mu_strerror (rc)));
-
-      /* Fallback to regular file stream */
-      rc = mu_file_stream_create (&mailbox->stream, dmp->name, mailbox->flags);
-      mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
-		("%s:%s (%s): %s",
-		 __func__, "mu_file_stream_create", dmp->name,
-		 mu_strerror (rc)));
-
-      if (rc)
-	return rc;
+      return rc;
     }
-
   mu_stream_set_buffer (mailbox->stream, mu_buffer_full, 0);
+  mu_stream_get_flags (mailbox->stream, &dmp->stream_flags);
   return 0;
 }
 
@@ -115,15 +99,15 @@ dotmail_open (mu_mailbox_t mailbox, int flags)
   mailbox->flags = flags;
 
   rc = dotmail_mailbox_init_stream (dmp);
-
-  if (mailbox->locker == NULL
+  if (rc == 0
+      && mailbox->locker == NULL
       && (flags & (MU_STREAM_WRITE | MU_STREAM_APPEND | MU_STREAM_CREAT)))
     {
-      rc = mu_locker_create (&mailbox->locker, dmp->name, 0);
+      rc = mu_locker_create_ext (&mailbox->locker, dmp->name, NULL);
       if (rc)
 	mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
 		  ("%s:%s (%s): %s",
-		   __func__, "mu_locker_create", dmp->name,
+		   __func__, "mu_locker_create_ext", dmp->name,
 		   mu_strerror (rc)));
     }
 
@@ -340,7 +324,7 @@ dotmail_message_alloc_uid (struct mu_dotmail_message *dmsg)
  * parameters, correspondingly.
  *
  * The header is stored in the first message.  To avoid rewriting entire
- * mailbox when one of the parameters chages, the values of <V> and <N>
+ * mailbox when one of the parameters changes, the values of <V> and <N>
  * are left-padded with spaces to the maximum width of their data types.
  *
  * Offset of the header in the mailbox and its length (without the
@@ -807,7 +791,8 @@ dotmail_message_unseen (mu_mailbox_t mailbox, size_t *pmsgno)
 	}
     }
 
-  return MU_ERR_NOENT;
+  *pmsgno = 0;
+  return 0;
 }
 
 /* Initialize the mailbox UID subsystem. See the Notes above. */
@@ -935,7 +920,8 @@ dotmail_quick_get_message (mu_mailbox_t mailbox, mu_message_qid_t qid,
 }
 
 static int
-mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg)
+mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg,
+			mu_envelope_t env, mu_attribute_t atr)
 {
   int rc;
   mu_off_t size;
@@ -943,6 +929,8 @@ mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg)
   static char *exclude_headers[] = {
     MU_HEADER_X_IMAPBASE,
     MU_HEADER_X_UID,
+    MU_HEADER_STATUS,
+    NULL,
     NULL
   };
   struct mu_dotmail_mailbox *dmp = mailbox->data;
@@ -957,10 +945,63 @@ mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg)
 
   do
     {
+      char statbuf[MU_STATUS_BUF_SIZE];
+      
+      if (atr)
+	{
+	  rc = mu_attribute_to_string (atr, statbuf, MU_STATUS_BUF_SIZE, NULL);
+	  if (rc)
+	    break;
+	}      
+      else
+	statbuf[0] = 0;
+      
+      if (env)
+	{
+	  char const *s;
+
+	  if (mu_envelope_sget_sender (env, &s) == 0)
+	    exclude_headers[3] = MU_HEADER_RETURN_PATH;
+	  if (mu_envelope_sget_date (env, &s) == 0)
+	    {
+	      struct tm tm;
+	      struct mu_timezone tz;
+	      
+	      if (mu_parse_date_dtl (s, NULL, NULL, &tm, &tz, NULL) == 0)
+		{
+		  /* Format a "Received:" header with that date */
+		  mu_stream_printf (mailbox->stream,
+				    "Received: from %s\n"
+				    "\tby %s; ",
+				    "localhost", "localhost");
+		  mu_c_streamftime (mailbox->stream,
+				    MU_DATETIME_FORM_RFC822, &tm, &tz);
+		  mu_stream_write (mailbox->stream, "\n", 1, NULL);
+		}
+	    }
+	}
+      
       rc = mu_stream_header_copy (mailbox->stream, istr, exclude_headers);
       if (rc)
 	break;
 
+      if (env)
+	{
+	  char const *s;
+
+	  if (mu_envelope_sget_sender (env, &s) == 0)
+	    {
+	      /* Create a Return-Path header */
+	      mu_stream_printf (mailbox->stream, "%s: %s\n",
+				MU_HEADER_RETURN_PATH, s);
+	    }
+	}
+      
+      /* Write status header */
+      if (statbuf[0])
+	mu_stream_printf (mailbox->stream,
+			  "%s: %s\n", MU_HEADER_STATUS, statbuf);      
+      
       /* Write UID-related data */
       if (dmp->uidvalidity_scanned)
 	{
@@ -1016,7 +1057,8 @@ mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg)
 }
 
 static int
-dotmail_append_message (mu_mailbox_t mailbox, mu_message_t msg)
+dotmail_append_message (mu_mailbox_t mailbox, mu_message_t msg,
+			mu_envelope_t env, mu_attribute_t atr)
 {
   struct mu_dotmail_mailbox *dmp = mailbox->data;
   int rc;
@@ -1035,7 +1077,7 @@ dotmail_append_message (mu_mailbox_t mailbox, mu_message_t msg)
     }
   else
     {
-      rc = mailbox_append_message (mailbox, msg);
+      rc = mailbox_append_message (mailbox, msg, env, atr);
 
       if (mailbox->locker)
 	mu_locker_unlock (mailbox->locker);
@@ -1077,18 +1119,63 @@ dotmail_get_size (mu_mailbox_t mailbox, mu_off_t *psize)
   return 0;
 }
 
+static int
+dotmail_stat (mu_mailbox_t mailbox, struct stat *st)
+{
+  int rc;
+  mu_transport_t trans[2];
+  
+  rc = mu_stream_ioctl (mailbox->stream, MU_IOCTL_TRANSPORT,
+			MU_IOCTL_OP_GET, trans);
+  if (rc == 0)
+    {
+      if (fstat ((int) (intptr_t) trans[0], st))
+	rc = errno;
+    }
+  return rc;
+}
+
+static int
+dotmail_set_priv (struct mu_dotmail_mailbox *dmp, struct stat *st)
+{
+  int rc;
+  mu_transport_t trans[2];
+  
+  rc = mu_stream_ioctl (dmp->mailbox->stream, MU_IOCTL_TRANSPORT,
+			MU_IOCTL_OP_GET, trans);
+  if (rc == 0)
+    {
+      int fd = (intptr_t) trans[0];
+      if (fchmod (fd, st->st_mode))
+	{
+	  mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+		    ("%s:%s: chmod failed: %s",
+		     __func__, dmp->name, strerror (errno)));
+	  rc = errno;
+	}
+      else if (fchown (fd, st->st_uid, st->st_gid))
+	{
+	  mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+		    ("%s:%s: chown failed: %s",
+		     __func__, dmp->name, strerror (errno)));
+	  rc = errno;
+	}
+    }
+  return rc;
+}
+
 static int
 dotmail_get_atime (mu_mailbox_t mailbox, time_t *return_time)
 {
   struct mu_dotmail_mailbox *dmp = mailbox->data;
   struct stat st;
-
+  int rc;
+  
   if (dmp == NULL)
     return EINVAL;
-  if (stat (dmp->name, &st))
-    return errno;
-  *return_time = st.st_atime;
-  return 0;
+  if ((rc = dotmail_stat (mailbox, &st)) == 0)
+    *return_time = st.st_atime;
+  return rc;
 }
 
 struct mu_dotmail_flush_tracker
@@ -1135,6 +1222,9 @@ dotmail_tracker_sync (struct mu_dotmail_flush_tracker *trk)
 	mu_dotmail_message_free (dmp->mesg[i]);
       dmp->size = 0;
       dmp->uidvalidity_scanned = 0;
+      dmp->uidvalidity_changed = 0;
+      dmp->x_imapbase_off = 0;
+      dmp->x_imapbase_len = 0;      
     }
   else
     {
@@ -1286,6 +1376,54 @@ dotmail_flush_temp (struct mu_dotmail_flush_tracker *trk,
   return mu_stream_flush (tempstr);
 }
 
+/*
+ * Copy the temporary mailbox stream TEMPSTR to the mailbox referred to by
+ * the tracker TRK.
+ */
+static inline int
+dotmail_copyback (struct mu_dotmail_flush_tracker *trk, mu_stream_t tempstr)
+{
+  int rc;
+  mu_stream_t mbx_stream = trk->dmp->mailbox->stream;
+  mu_off_t size;
+  
+  rc = mu_stream_seek (tempstr, 0, MU_SEEK_SET, NULL);
+  if (rc)
+    {
+      mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+		("%s: can't rewind temporary file: %s",
+		 __func__, mu_strerror (rc)));
+      return rc;
+    }
+
+  rc = mu_stream_seek (mbx_stream, 0, MU_SEEK_SET, NULL);
+  if (rc)
+    {
+      mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+		("%s: can't rewind mailbox %s: %s",
+		 __func__, trk->dmp->name, mu_strerror (rc)));
+      return rc;
+    }
+
+  rc = mu_stream_copy (mbx_stream, tempstr, 0, &size);
+  if (rc)
+    {
+      mu_error (_("copying back to mailbox %s failed: %s"),
+		trk->dmp->name, mu_strerror (rc));
+      return rc;
+    }
+  rc = mu_stream_truncate (mbx_stream, size);
+  if (rc)
+    {
+      mu_error (_("cannot truncate mailbox stream: %s"),
+		mu_stream_strerror (mbx_stream, rc));
+      return rc;
+    }
+  
+  dotmail_tracker_sync (trk);
+  return 0;
+}
+
 /* Flush the mailbox described by the tracker TRK to the stream TEMPSTR.
    EXPUNGE is 1 if the MU_ATTRIBUTE_DELETED attribute is to be honored.
    Assumes that simultaneous access to the mailbox has been blocked.
@@ -1351,127 +1489,170 @@ dotmail_flush_unlocked (struct mu_dotmail_flush_tracker *trk, int mode)
 			("%s:%s (%s): %s",
 			 __func__, "mu_stream_printf", dmp->name,
 			 mu_strerror (rc)));
+	      
 	    }
 	  return 0;
 	}
-
-      /*
-       * There is no X-IMAPbase header yet or it is not wide enough to
-       * accept the current value.  Fall back to reformatting entire
-       * mailbox.  Clear any other changes that might have been done
-       * to its messages.
-       */
-      dirty = 0;
-      dmp->mesg[0]->uid_modified = 1;
-
-      for (i = 1; i < dmp->mesg_count; i++)
+      else
 	{
-	  struct mu_dotmail_message *dmsg = dmp->mesg[i];
-	  dmsg->uid_modified = 0;
-	  dmsg->attr_flags &= ~(MU_ATTRIBUTE_MODIFIED|MU_ATTRIBUTE_DELETED);
-	}
-    }
-  else
-    {
-      for (dirty = 0; dirty < dmp->mesg_count; dirty++)
-	{
-	  struct mu_dotmail_message *dmsg = dmp->mesg[dirty];
-	  if (dmsg->uid_modified)
-	    break;
-	  mu_dotmail_message_attr_load (dmsg);
-	  if ((dmsg->attr_flags & MU_ATTRIBUTE_MODIFIED)
-	      || (dmsg->attr_flags & MU_ATTRIBUTE_DELETED)
-	      || (dmsg->message && mu_message_is_modified (dmsg->message)))
-	    break;
+	  /*
+	   * There is no X-IMAPbase header yet or it is not wide enough to
+	   * accept the current value.  Fall back to reformatting entire
+	   * mailbox.  Clear any other changes that might have been done
+	   * to its messages.
+	   */
+	  dirty = 0;
+	  dmp->mesg[0]->uid_modified = 1;
+	  
+	  for (i = 1; i < dmp->mesg_count; i++)
+	    {
+	      struct mu_dotmail_message *dmsg = dmp->mesg[i];
+	      dmsg->uid_modified = 0;
+	      dmsg->attr_flags &= ~(MU_ATTRIBUTE_MODIFIED|MU_ATTRIBUTE_DELETED);
+	    }
 	}
     }
   
-  p = strrchr (dmp->name, '/');
-  if (p)
+  for (dirty = 0; dirty < dmp->mesg_count; dirty++)
     {
-      size_t l = p - dmp->name;
-      hints.tmpdir = malloc (l + 1);
-      if (!hints.tmpdir)
-	return ENOMEM;
-      memcpy (hints.tmpdir, dmp->name, l);
-      hints.tmpdir[l] = 0;
-    }
-  else
-    {
-      hints.tmpdir = mu_getcwd ();
-      if (!hints.tmpdir)
-	return ENOMEM;
-    }
-  rc = mu_tempfile (&hints, MU_TEMPFILE_TMPDIR, &tempfd, &tempname);
-  if (rc)
-    {
-      free (hints.tmpdir);
-      return rc;
-    }
-  rc = mu_fd_stream_create (&tempstr, tempname, tempfd,
-			    MU_STREAM_RDWR|MU_STREAM_SEEK);
-  if (rc)
-    {
-      free (hints.tmpdir);
-      close (tempfd);
-      free (tempname);
-      return rc;
+      struct mu_dotmail_message *dmsg = dmp->mesg[dirty];
+      if (dmsg->uid_modified)
+	break;
+      mu_dotmail_message_attr_load (dmsg);
+      if ((dmsg->attr_flags & MU_ATTRIBUTE_MODIFIED)
+	  || (dmsg->attr_flags & MU_ATTRIBUTE_DELETED)
+	  || (dmsg->message && mu_message_is_modified (dmsg->message)))
+	break;
     }
 
-  rc = dotmail_flush_temp (trk, dirty, tempstr, mode == FLUSH_EXPUNGE);
-  mu_stream_unref (tempstr);
-  if (rc == 0)
-    {
-      /* Rename mailbox to temporary copy */
-      char *backup = mu_tempname (hints.tmpdir);
-      rc = rename (dmp->name, backup);
-      if (rc)
+  rc = 0;
+  if (dirty < dmp->mesg_count)
+    {  
+      p = strrchr (dmp->name, '/');
+      if (p)
 	{
-	  mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
-		("%s:%s: failed to rename to backup file %s: %s",
-		 __func__, dmp->name, tempname,
-		 mu_strerror (rc)));
-	  unlink (backup);
+	  size_t l = p - dmp->name;
+	  hints.tmpdir = malloc (l + 1);
+	  if (!hints.tmpdir)
+	    return ENOMEM;
+	  memcpy (hints.tmpdir, dmp->name, l);
+	  hints.tmpdir[l] = 0;
 	}
       else
 	{
-	  rc = rename (tempname, dmp->name);
-	  if (rc == 0)
+	  hints.tmpdir = mu_getcwd ();
+	  if (!hints.tmpdir)
+	    return ENOMEM;
+	}
+      rc = mu_tempfile (&hints, MU_TEMPFILE_TMPDIR, &tempfd, &tempname);
+      if (rc == 0)
+	{
+	  rc = mu_fd_stream_create (&tempstr, tempname, tempfd,
+				    MU_STREAM_RDWR|MU_STREAM_SEEK);
+	}
+      else if (rc == EACCES)
+	{
+	  /*
+	   * Mail spool directory is not writable for the user. Fall
+	   * back to using temporary stream located elsewhere. When
+	   * ready, it will be copied back to the mailbox.
+	   *
+	   * Reset the tempname to NULL to instruct the code below
+	   * which approach to take.
+	   */
+	  tempname = NULL;
+	  
+	  rc = mu_temp_file_stream_create (&tempstr, NULL, 0);
+	}
+      
+      if (rc)
+	{
+	  free (hints.tmpdir);
+	  close (tempfd);
+	  free (tempname);
+	  return rc;
+	}
+      
+      rc = dotmail_flush_temp (trk, dirty, tempstr, mode == FLUSH_EXPUNGE);
+      if (rc == 0)
+	{
+	  if (tempname)
 	    {
-	      /* Success. Synchronize internal data with the counter. */
-	      dotmail_tracker_sync (trk);
-	      mu_stream_destroy (&dmp->mailbox->stream);
-	      rc = dotmail_mailbox_init_stream (dmp);
+	      /* Mail spool is writable. Rename the temporary copy back
+		 to mailbox */
+	      char *backup;
+	      struct stat st;
+	      
+	      if ((rc = dotmail_stat (dmp->mailbox, &st)) != 0)
+		{
+		  mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+			    ("%s:%s: stat failed: %s",
+			     __func__, dmp->name, strerror (errno)));
+		}
+	      else
+		{
+		  mu_stream_flush (tempstr);
+		  backup = mu_tempname (hints.tmpdir);
+		  if (rename (dmp->name, backup))
+		    {
+		      rc = errno;
+		      mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+				("%s:%s: failed to rename to backup file %s: %s",
+				 __func__, dmp->name, tempname,
+				 mu_strerror (rc)));
+		      unlink (backup);
+		    }
+		  else
+		    {
+		      rc = rename (tempname, dmp->name);
+		      if (rc == 0)
+			{
+			  /* Success. Synchronize internal data with the
+			     counter. */
+			  dotmail_tracker_sync (trk);
+			  mu_stream_destroy (&dmp->mailbox->stream);
+			  rc = dotmail_mailbox_init_stream (dmp);
+			  if (rc == 0)
+			    dotmail_set_priv (dmp, &st);
+			}
+		      else
+			{
+			  int rc1;
+			  mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+				    ("%s: failed to rename temporary file %s %s: %s",
+				     __func__, tempname, dmp->name,
+				     mu_strerror (rc)));
+			  rc1 = rename (backup, dmp->name);
+			  if (rc1)
+			    {
+			      mu_error (_("failed to restore %s from backup %s: %s"),
+					dmp->name, backup, mu_strerror (rc1));
+			      mu_error (_("backup left in %s"), backup);
+			      free (backup);
+			      backup = NULL;
+			    }
+			}
+		    }
+		  
+		  if (backup)
+		    {
+		      unlink (backup);
+		      free (backup);
+		    }
+		  unlink (tempname);
+		}
 	    }
 	  else
 	    {
-	      int rc1;
-	      mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
-			("%s: failed to rename temporary file %s %s: %s",
-			 __func__, tempname, dmp->name,
-			 mu_strerror (rc)));
-	      rc1 = rename (backup, dmp->name);
-	      if (rc1)
-		{
-		  mu_error (_("failed to restore %s from backup %s: %s"),
-			    dmp->name, backup, mu_strerror (rc1));
-		  mu_error (_("backup left in %s"), backup);
-		  free (backup);
-		  backup = NULL;
-		}
-	    }
-	}
-
-      if (backup)
-	{
-	  unlink (backup);
-	  free (backup);
+	      /* Mail spool not writable.  Copy the tempstr back to mailbox. */
+	      rc = dotmail_copyback (trk, tempstr);
+	    }	    
 	}
+      free (tempname);
+      free (hints.tmpdir);
+      mu_stream_unref (tempstr);
     }
-  unlink (tempname);
-  free (tempname);
-  free (hints.tmpdir);
-
+  
   dmp->uidvalidity_changed = 0;  
   
   return rc;
