@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2009-2020 Free Software Foundation, Inc.
+   Copyright (C) 2009-2024 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -52,7 +52,7 @@ show_struct (msgset_t *msgset, mu_message_t msg, void *data)
   mime_descend (&mclos, show_part, NULL);
 
     /* Mark enclosing message as read */
-  if (mu_mailbox_get_message (mbox, msgset->msg_part[0], &msg) == 0)
+  if (mu_mailbox_get_message (mbox, msgset_msgno (msgset), &msg) == 0)
     util_mark_read (msg);
 
   return 0;
@@ -61,6 +61,7 @@ show_struct (msgset_t *msgset, mu_message_t msg, void *data)
 int
 mail_struct (int argc, char **argv)
 {
-  return util_foreach_msg (argc, argv, MSG_NODELETED|MSG_SILENT,
+  return util_foreach_msg (argc, argv,
+			   MSG_NODELETED|MSG_SILENT,
 			   show_struct, NULL);
 }
