@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2013-2022 Free Software Foundation, Inc.
+   Copyright (C) 2013-2024 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -274,12 +274,18 @@ checkmail (const char *username, int personal)
 	      break;
 
 	    default:
-	      if (rc != ENOSYS && rc != MU_ERR_INFO_UNAVAILABLE)
-		mu_diag_funcall (MU_DIAG_ERROR, "mu_mailbox_messages_unseen",
-				 mu_url_to_string (url), rc);
+	      mu_diag_funcall (MU_DIAG_ERROR, "mu_mailbox_messages_unseen",
+			       mu_url_to_string (url), rc);
+	      /* FALLTHROUGH */
+	    case EAGAIN:
+	    case EINTR:
+	    case ENOSYS:
+	    case EINPROGRESS:
+	    case MU_ERR_INFO_UNAVAILABLE:
 	      rc = mu_mailbox_messages_recent (mbox, &recent);
 	      if (rc == 0)
 		newmail = recent > 0;
+	      break;
 	    }
 
 	  if (rc)
