@@ -1,5 +1,5 @@
 #  GNU Mailutils -- a suite of utilities for electronic mail
-#  Copyright (C) 2009-2012, 2014-2016 Free Software Foundation, Inc.
+#  Copyright (C) 2009-2024 Free Software Foundation, Inc.
 #
 #  This library is free software; you can redistribute it and/or
 #  modify it under the terms of the GNU Lesser General Public
@@ -15,7 +15,6 @@
 #  Public License along with this library.  If not, see
 #  <http://www.gnu.org/licenses/>.
 
-import types
 from mailutils.c_api import header
 from mailutils.error import *
 
@@ -27,7 +26,7 @@ class Header:
         del self.hdr
 
     def __getitem__ (self, item):
-        if isinstance (item, types.IntType):
+        if isinstance(item, int):
             return self.get_field_name (item), self.get_field_value (item)
         else:
             return self.get_value (item)
@@ -49,7 +48,7 @@ class Header:
         elif name == 'lines':
             return self.get_lines ()
         else:
-            raise AttributeError, name
+            raise AttributeError(name)
 
     def __len__ (self):
         return self.get_field_count ()
@@ -59,7 +58,7 @@ class Header:
         self.__len = self.get_field_count ()
         return self
 
-    def next (self):
+    def __next__ (self):
         if self.__count >= self.__len:
             self.__count = 0
             raise StopIteration
@@ -177,6 +176,7 @@ MU_HEADER_FCC =                       "Fcc"
 MU_HEADER_DELIVERY_DATE =             "Delivery-date"
 MU_HEADER_ENVELOPE_TO =               "Envelope-to"
 MU_HEADER_X_EXPIRE_TIMESTAMP =        "X-Expire-Timestamp"
+MU_HEADER_USER_AGENT =                "User-Agent"
 
 MU_HEADER_REPLACE = 0x01
 MU_HEADER_BEFORE  = 0x02
